"""Flask API."""

import json
import logging
import time
import traceback
import typing
from http import HTTPStatus

from flask import Flask, Response, jsonify, render_template, request
from werkzeug.exceptions import BadRequest

from ripple1d.api import tasks
from ripple1d.api.utils import get_unexpected_and_missing_args
from ripple1d.ops.fim_lib import create_fim_lib, create_rating_curves_db, fim_lib_stac, nwm_reach_model_stac
from ripple1d.ops.metrics import compute_conflation_metrics
from ripple1d.ops.ras_conflate import conflate_model
from ripple1d.ops.ras_run import (
    create_model_run_normal_depth,
    run_incremental_normal_depth,
    run_known_wse,
)
from ripple1d.ops.ras_terrain import create_ras_terrain
from ripple1d.ops.subset_gpkg import extract_submodel
from ripple1d.ras_to_gpkg import gpkg_from_ras

app = Flask(__name__)


@app.route("/processes/conflate_model/execution", methods=["POST"])
def process__conflate_model():
    """Enqueue a task to conflate a source model."""
    return enqueue_async_task(conflate_model)


@app.route("/processes/compute_conflation_metrics/execution", methods=["POST"])
def process__compute_conflation_metrics():
    """Enqueue a task to compute conflation metrics."""
    return enqueue_async_task(compute_conflation_metrics)


@app.route("/processes/gpkg_from_ras/execution", methods=["POST"])
def process__gpkg_from_ras():
    """Enqueue a task to create a new GeoPackage from a source model."""
    return enqueue_async_task(gpkg_from_ras)


@app.route("/processes/extract_submodel/execution", methods=["POST"])
def process__extract_submodel():
    """Enqueue a task to create a new GeoPackage for a NWM reach."""
    return enqueue_async_task(extract_submodel)


@app.route("/processes/create_ras_terrain/execution", methods=["POST"])
def process__create_ras_terrain():
    """Enqueue a task to create a new RAS terrain."""
    return enqueue_async_task(create_ras_terrain)


@app.route("/processes/create_model_run_normal_depth/execution", methods=["POST"])
def process__create_model_run_normal_depth():
    """Enqueue a task to calculate the initial normal depth."""
    return enqueue_async_task(create_model_run_normal_depth)


@app.route("/processes/run_incremental_normal_depth/execution", methods=["POST"])
def process__run_incremental_normal_depth():
    """Enqueue a task to calculate the incremental normal depth."""
    return enqueue_async_task(run_incremental_normal_depth)


@app.route("/processes/run_known_wse/execution", methods=["POST"])
def process__run_known_wse():
    """Enqueue a task to calculate the water surface elevation (WSE) based on known inputs."""
    return enqueue_async_task(run_known_wse)


@app.route("/processes/create_rating_curves_db/execution", methods=["POST"])
def process__create_rating_curves_db():
    """Enqueue a task to create a rating curve db."""
    return enqueue_async_task(create_rating_curves_db)


@app.route("/processes/create_fim_lib/execution", methods=["POST"])
def process__create_fim_lib():
    """Enqueue a task to create a FIM library."""
    return enqueue_async_task(create_fim_lib)


@app.route("/processes/nwm_reach_model_stac/execution", methods=["POST"])
def process__nwm_reach_model_stac():
    """Enqueue a task to create a stac item from a fim model."""
    return enqueue_async_task(nwm_reach_model_stac)


@app.route("/processes/fim_lib_stac/execution", methods=["POST"])
def process__fim_lib_stac():
    """Enqueue a task to create a stac item from a fim library."""
    return enqueue_async_task(fim_lib_stac)


@app.route("/ping", methods=["GET"])
def ping():
    """Check the health of the service."""
    return jsonify({"status": "healthy"}), HTTPStatus.OK


@app.route("/processes/test/execution", methods=["POST"])
def test():
    """Test the execution and monitoring of an asynchronous task."""
    response, http_status = enqueue_async_task(tasks.noop)
    if http_status != HTTPStatus.CREATED:
        return jsonify(response.json, HTTPStatus.INTERNAL_SERVER_ERROR)

    timeout_seconds = 10

    start = time.time()
    while time.time() - start < timeout_seconds:
        time.sleep(0.2)
        status = tasks.fetch_ogc_status(response.json["jobID"])
        if status == "failed":
            return jsonify({"status": "not healthy"}), HTTPStatus.INTERNAL_SERVER_ERROR
        if status == "successful":
            return jsonify({"status": "healthy"}), HTTPStatus.OK
    return (
        jsonify({"status": f"huey is busy or not active, ping timed out after {timeout_seconds} seconds"}),
        HTTPStatus.GATEWAY_TIMEOUT,
    )


@app.route("/jobs", methods=["GET"])
def jobs():
    """Retrieve OGC status and result for all jobs."""
    # try:
    format_option = request.args.get("f")
    task2metadata = tasks.task_status(only_task_id=None)
    jobs = [get_job_status(task_id, huey_metadata) for task_id, huey_metadata in task2metadata.items()]
    response = {"jobs": jobs}

    if format_option == "json":
        return jsonify(response), HTTPStatus.OK
    else:
        return render_template("jobs.html", response=response)


@app.route("/jobs/<task_id>", methods=["GET"])
def job_status(task_id):
    """Retrieve result for job."""
    include_traceback, problem = parse_request_param__bool(param_name="tb", default=False)
    if problem is not None:
        return problem
    task2metadata = tasks.task_summary(only_task_id=task_id)
    resp = get_job_status(task_id, task2metadata[task_id], return_result=True, include_traceback=include_traceback)
    try:
        return jsonify(resp), HTTPStatus.OK

    except Exception as e:
        return (
            jsonify(str(e)),
            HTTPStatus.NOT_FOUND,
        )


@app.route("/jobs/<task_id>/logs", methods=["GET"])
def job_logs(task_id):
    """Retrieve result for job."""
    try:
        result = tasks.fetch_logs(task_id)
        return jsonify(result), HTTPStatus.OK
    except Exception as e:
        return (
            jsonify(str(e)),
            HTTPStatus.NOT_FOUND,
        )


@app.route("/jobs/<task_id>/results", methods=["GET"])
def job_results(task_id):
    """Retrieve result for job."""
    try:
        result = tasks.fetch_results(task_id)
        return jsonify(result), HTTPStatus.OK
    except Exception as e:
        return (
            jsonify(str(e)),
            HTTPStatus.NOT_FOUND,
        )


@app.route("/jobs/<task_id>/metadata", methods=["GET"])
def job_metadata(task_id):
    """Retrieve metadata for job."""
    task2metadata = tasks.task_status(only_task_id=task_id)
    try:
        return jsonify(task2metadata), HTTPStatus.OK

    except Exception as e:
        return (
            jsonify(str(e)),
            HTTPStatus.NOT_FOUND,
        )


@app.route("/jobs/<task_id>", methods=["DELETE"])
def dismiss(task_id):
    """Dismiss a specific task by its ID.

    Available only for jobs in "accepted" status, dismissal of "running" jobs not implemented
    """
    try:
        ogc_status = tasks.fetch_ogc_status(task_id)
        if ogc_status == "notfound":
            return jsonify({"type": "process", "detail": f"job ID not found: {task_id}"}), HTTPStatus.NOT_FOUND

        elif ogc_status == "accepted":
            tasks.revoke_task(task_id)
            return jsonify({"type": "process", "detail": f"job ID dismissed: {task_id}"}), HTTPStatus.OK

        elif ogc_status == "running":
            tasks.revoke_task_by_pid(task_id)
            return jsonify({"type": "process", "detail": f"job ID dismissed: {task_id}"}), HTTPStatus.OK

        elif ogc_status == "dismissed":
            return jsonify({"type": "process", "detail": f"job ID dismissed: {task_id}"}), HTTPStatus.OK

        else:
            return (
                jsonify(
                    {
                        "type": "process",
                        "detail": f"failed to dismiss job ID {task_id} due to job status '{ogc_status}'",
                    }
                ),
                HTTPStatus.CONFLICT,
            )
    except Exception as e:
        return (
            jsonify(
                {
                    "type": "process",
                    "detail": f"failed to dismiss job ID {task_id} due to internal server error {str(e)}",
                }
            ),
            HTTPStatus.INTERNAL_SERVER_ERROR,
        )


def parse_request_param__bool(param_name: str, default: bool) -> tuple[bool, tuple]:
    """Get the parameter, assert it is true or false, and return the appropriate Python boolean value as well as a status tuple.

    If there is a problem, the status tuple has two elements: a response message, and a HTTP status to be returned by the endpoint.
    If there is not a problem, the status tuple is actually None.
    """
    arg_tb = request.args.get(param_name)
    if not arg_tb:
        return (default, None)
    elif arg_tb.lower() == "false":
        return (False, None)
    elif arg_tb.lower() == "true":
        return (True, None)
    else:
        return (
            None,
            (
                jsonify(
                    {
                        "type": "process",
                        "detail": f"query param 'tb' should be 'true' or 'false', but got: {repr(arg_tb)}",
                    }
                ),
                HTTPStatus.BAD_REQUEST,
            ),
        )


def get_job_status(task_id: str, huey_metadata: dict, return_result: bool = False, include_traceback: bool = False) -> dict:
    """Convert huey-style task status metadata into a OGC-style job summary dictionary."""
    out_dict = {
        "jobID": task_id,
        "updated": huey_metadata["status_time"],
        "status": huey_metadata["ogc_status"],
        "processID": huey_metadata["func_name"],
    }
    if return_result:
        if not include_traceback and huey_metadata['result'] is not None:
            del huey_metadata['result']['tb']
        out_dict["result"] = huey_metadata['result']
    return out_dict


def enqueue_async_task(func: typing.Callable) -> tuple[Response, HTTPStatus]:
    """Start the execution of the provided func using kwargs from the request body. Assume body is a JSON dictionary."""
    try:
        kwargs = request.json  # can throw BadRequest when parsing body into json
        if not isinstance(kwargs, dict):
            raise BadRequest(f"expected body to be a JSON dictionary, but got: {type(kwargs)}")
    except BadRequest as e:
        return (
            jsonify({"type": "process", "detail": f"could not parse body to json dict. error: {e}"}),
            HTTPStatus.BAD_REQUEST,
        )

    unexpected, missing = get_unexpected_and_missing_args(func, set(kwargs))
    if unexpected or missing:
        return (
            jsonify({"type": "process", "detail": f"unexpected args: {unexpected}, missing args: {missing}"}),
            HTTPStatus.BAD_REQUEST,
        )

    try:
        result = tasks.create_and_enqueue_task(func, kwargs)
    except:
        msg = f"unable to submit task: {traceback.format_exc()}"
        app.logger.error(msg)
        return jsonify({"type": "process", "detail": msg}), HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return (
            jsonify({"type": "process", "jobID": result.id, "status": "accepted"}),
            HTTPStatus.CREATED,
        )



"""Utils for working with geopackages."""

import json
import logging
import os
import re
import subprocess
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import Dict

import boto3
import contextily as ctx
import fiona
import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
import pystac
import requests
from pyproj import CRS
from shapely import Polygon, to_geojson

from ripple1d.consts import LAYER_COLORS
from ripple1d.data_model import NwmReachModel, RasModelStructure
from ripple1d.errors import UnkownCRSUnitsError
from ripple1d.ras import (
    VALID_GEOMS,
    VALID_PLANS,
    VALID_STEADY_FLOWS,
    RasFlowText,
    RasGeomText,
    RasManager,
    RasPlanText,
    RasProject,
)
from ripple1d.utils.s3_utils import get_basic_object_metadata, init_s3_resources, str_from_s3


def get_river_miles(river_gdf: gpd.GeoDataFrame):
    """Compute the total length of the river centerlines in miles."""
    if "units" not in river_gdf.crs.to_dict().keys():
        raise UnkownCRSUnitsError("No units specified. The coordinate system may be Geographic.")
    units = river_gdf.crs.to_dict()["units"]
    if units in ["ft-us", "ft", "us-ft"]:
        conversion_factor = 1 / 5280
    elif units in ["m", "meters"]:
        conversion_factor = 1 / 1609
    else:
        raise UnkownCRSUnitsError(f"Expected feet or meters; got: {units}")
    return round(river_gdf.length.sum() * conversion_factor, 2)


def gpkg_to_geodataframe(gpkg_s3_uri: str) -> dict:
    """
    Convert a local geopackage file to a GeoDataFrame.

    Parameters
    ----------
        gpkg_key (str): Path of locally saved geopackage.

    Returns
    -------
        gpkg_gdf (dict): dictionary of GeoDataFrame.

    """
    layers = fiona.listlayers(gpkg_s3_uri)
    gdfs = {}
    for layer in layers:
        gdfs[layer] = gpd.read_file(gpkg_s3_uri, layer=layer)

    return gdfs


def reproject(gdfs: dict, crs=4326) -> dict:
    """Reproject a gdf to a new CRS."""
    for layer, gdf in gdfs.items():
        gdfs[layer] = gdf.to_crs(crs)  # reproject to WSG 84
    return gdfs


def create_thumbnail_from_gpkg(gdfs: dict) -> plt.Figure:
    """
    Create a figure displaying the geopandas dataframe provided in the gdfs dictionary.

    Parameters
    ----------
    - gdf (dict): A dictionary of geopandas dataframes containing the geometries to plot.
    """
    # Define colors for each layer type

    # Plotting
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))

    for layer, color in LAYER_COLORS.items():
        if layer in gdfs.keys():
            gdfs[layer].plot(ax=ax, color=color, linewidth=1, label=layer)
            crs = gdfs[layer].crs

    # Add openstreetmap basemap
    try:
        ctx.add_basemap(ax, crs=crs, source=ctx.providers.USGS.USTopo)
    except requests.exceptions.HTTPError as e:
        try:
            ctx.add_basemap(ax, crs=crs, source=ctx.providers.Esri.WorldStreetMap)
        except requests.exceptions.HTTPError as e:
            ctx.add_basemap(ax, crs=crs, source=ctx.providers.OpenStreetMap.Mapnik)
    ax.legend()

    # Hide all axis text ticks or tick labels
    ax.set_xticks([])
    ax.set_yticks([])
    return fig


def write_thumbnail_to_s3(fig: plt.Figure, png_s3_key: str, bucket: str, s3_client: boto3.Session.client):
    """
    Write a PNG thumbnail to AWS S3.

    Parameters
    ----------
    - fig (plt.Figure): The figure to export to png on s3.
        - png_s3_key (str): The S3 path where the generated PNG thumbnail is to be stored.
    - bucket (str): The S3 bucket
    - s3_client: The AWS S3 client instance used for uploading the PNG.
    """
    # Save plot to a bytes buffer
    buf = BytesIO()
    plt.savefig(buf, format="png")
    buf.seek(0)
    plt.close()

    # Download the PNG to s3
    s3_client.put_object(Bucket=bucket, Key=png_s3_key, Body=buf, ContentType="image/png")


def create_geom_item(
    item_id: str,
    bbox: Polygon,
    footprint: list[float],
    properties: dict,
) -> pystac.Item:
    """
    Create a PySTAC Item for a gpkg file stored in an AWS S3 bucket.

    Parameters
    ----------
        item_id (str): item_id.
        bbox (shapely.Polygon): Item bounding box.
        footprint (list[float]): Item Footprint.

    Returns
    -------
        pystac.Item: The PySTAC Item representing the gpkg file.
    """
    # TODO: Adjust start_time selection for item
    start_time = datetime.utcnow()
    item = pystac.Item(
        id=item_id,
        geometry=json.loads(to_geojson(footprint)),
        bbox=bbox.tolist(),
        datetime=start_time,
        properties=properties,
    )

    return item


def get_asset_info(asset_key: str, rms: RasModelStructure, bucket: str = None) -> dict:
    """
    Generate information for an asset based on its file extension.

    Parameters
    ----------
        asset_key (str): The S3 key of the asset.
        rms (str): The RasModelStructure.
        bucket (str): The S3 bucket where the asset is stored.

    Returns
    -------
        dict: A dictionary with the roles, the description, and the title of the asset.
    """
    if bucket:
        _, client, s3_resource = init_s3_resources()
        obj = s3_resource.Bucket(bucket).Object(asset_key)
        extra_fields = get_basic_object_metadata(obj)
    else:
        extra_fields = {"file:size": os.path.getsize(asset_key), "last_modified": os.path.getmtime(asset_key)}

    file_extension = Path(asset_key).suffix
    title = Path(asset_key).name.replace(" ", "_")
    if file_extension in VALID_PLANS + VALID_GEOMS + VALID_STEADY_FLOWS:
        if bucket:
            string = str_from_s3(rms.ras_project_file, client, bucket)
            rp = RasProject.from_str(string, rms.ras_project_file)
        else:
            rm = RasManager(rms.ras_project_file, crs=rms.crs)

    description = ""
    roles = []

    if re.match(".f[0-9]{2}", file_extension):
        roles.extend(["forcing", "hec-ras", pystac.MediaType.TEXT])
        description = """Forcing file for ras."""
        if bucket:
            string = str_from_s3(asset_key, client, bucket)
            flow = RasFlowText.from_str(string, " .f01")
        else:
            flow = [i for i in rm.flows.values() if i.file_extension == file_extension][0]
        extra_fields["Title"] = flow.title
        extra_fields["Number of Profiles"] = flow.n_profiles
        extra_fields["Profile Names"] = flow.profile_names

    elif re.match(".g[0-9]{2}", file_extension):
        roles.extend(["geometry", "hec-ras", pystac.MediaType.TEXT])
        description = """Geometry file for ras."""
        if bucket:
            string = str_from_s3(asset_key, client, bucket)
            geom = RasGeomText.from_str(string, rms.crs, " .g01")
        else:
            geom = [i for i in rm.geoms.values() if i.file_extension == file_extension][0]
        extra_fields["Title"] = geom.title
        extra_fields["Number of rivers"] = geom.n_rivers
        extra_fields["Number of reaches"] = geom.n_reaches
        extra_fields["Number of cross sections"] = geom.n_cross_sections
        extra_fields["Number of junctions"] = geom.n_junctions

    elif re.match(".p[0-9]{2}", file_extension):
        roles.extend(["plan", "hec-ras", pystac.MediaType.TEXT])
        description = """Plan file for ras."""
        if bucket:
            string = str_from_s3(asset_key, client, bucket)
            plan = RasPlanText.from_str(string, rms.crs, " .p01")
        else:
            plan = [i for i in rm.plans.values() if i.file_extension == file_extension][0]
        extra_fields["Title"] = plan.title
        # extra_fields["Geometry Title"] = plan.geom.title
        # extra_fields["Geometry Extension"] = plan.geom.file_extension
        # extra_fields["Flow Title"] = plan.flow.title
        # extra_fields["Flow Extension"] = plan.flow.file_extension

    elif re.match(".O[0-9]{2}", file_extension):
        roles.extend(["output", "hec-ras", pystac.MediaType.TEXT])
        description = """Output file for ras."""

    elif re.match(".r[0-9]{2}", file_extension):
        roles.extend(["run", "hec-ras", pystac.MediaType.TEXT])
        description = """Run file for ras."""

    elif re.match(".rasmap", file_extension):
        roles.extend(["rasmap", "hec-ras", pystac.MediaType.XML])
        description = """Rasmapper file for ras."""

    elif ".rasmap.backup" in title:
        roles.extend(["rasmap-backup", "hec-ras", pystac.MediaType.XML])
        description = """Rasmapper backup file for ras."""

    elif ".computeMsgs.txt" in title:
        roles.extend(["compute-messages", "hec-ras", pystac.MediaType.TEXT])
        description = """Compute messages file for ras."""

    elif file_extension == ".png":
        roles.extend(["thumbnail", pystac.MediaType.PNG])
        description = """PNG of geometry with OpenStreetMap basemap."""
        title = "Thumbnail"

    elif re.match(".gpkg", file_extension):
        roles.extend(["ras-geometry-gpkg", pystac.MediaType.GEOPACKAGE])
        description = """GeoPackage file with geometry data extracted from .gxx file."""
        title = title

    elif ".ripple1d.json" in title:
        roles.extend(["ripple-parameters", pystac.MediaType.JSON])
        description = """Json file containing Ripple parameters."""
        title = "Ripple parameters"

    elif ".stac.json" in title:
        roles.extend(["stac-item", pystac.MediaType.JSON])
        description = """Json for this stac-item."""
        title = title

    elif file_extension == ".prj":
        if bucket:
            _, client, _ = init_s3_resources()
            string = str_from_s3(asset_key, client, bucket)
        else:
            with open(asset_key, "r") as file:
                string = file.read()
        if "Proj Title=" in string:
            roles.extend(["project-file", "hec-ras", pystac.MediaType.TEXT])
            description = """Project file for ras."""
        else:
            if isinstance(rms, NwmReachModel):
                if asset_key in rms.terrain_assets:
                    roles.extend(["Terrain", "projection-file", pystac.MediaType.TEXT])
                else:
                    roles.extend(["projection-file", pystac.MediaType.TEXT])
                description = """Projection file for ras."""

    if isinstance(rms, NwmReachModel):
        if file_extension == ".hdf":
            if asset_key in rms.terrain_assets:
                roles.extend(["Terrain", pystac.MediaType.HDF])
            else:
                roles.extend(["hec-ras", pystac.MediaType.HDF])

        elif file_extension == ".vrt":
            if asset_key in rms.terrain_assets:
                roles.extend(["Terrain", pystac.MediaType.XML])

        elif file_extension == ".tif":
            if asset_key in rms.terrain_assets:
                roles.extend(["Terrain", pystac.MediaType.GEOTIFF])

    return {"roles": roles, "description": description, "title": title, "extra_fields": extra_fields}


def find_hash(item_metadata: Dict, asset_file: str) -> dict:
    """
    Extract the hash value for a given asset file from a metadata dictionary.

    This function searches through a metadata dictionary for an asset file based on the file's extension.
    It then extracts and returns the hash value associated with that file extension from the metadata.

    Parameters
    ----------
        item_metadata (dict): A dictionary containing metadata items with hash info.
        asset_file (str): The path to the asset file.

    Returns
    -------
        hash_dict (Dict): A dictionary with a single key-value pair where the value is the hash for the
        asset file. Returns an empty dictionary if no hash is found for the file extension.
    """
    hash_dict = {}

    file_extension = Path(asset_file).suffix.lstrip(".")
    for key in item_metadata.keys():
        key_file_ext = key.split(":")[-1]
        if file_extension == key_file_ext:
            hash_dict = {"hash": item_metadata[key]["Hash"]}

    return hash_dict


def remove_hash_from_metadata(item_metadata: Dict) -> dict:
    """
    Remove "Hash" from each key (if it exists) in metedata dictionary.

    Parameters
    ----------
        item_metadata (Dict): Dictionary of item metadata with hashs.

    Returns
    -------
        no_hash_metadata (Dict): New metadata dictionary without hash info.
    """
    # Copy the dictionary to avoid modifying the original
    no_hash_metadata = item_metadata.copy()
    for key in no_hash_metadata:
        if "Hash" in no_hash_metadata[key]:
            del no_hash_metadata[key]["Hash"]
    return no_hash_metadata
